import sys

from .pampy import (ANY, HEAD, REST, TAIL, MatchError, match,  # no _
                    match_dict, match_iterable, match_value)

__version__ = '0.3.0'


__all__ = [
    "match",
    "ANY",
    "HEAD",
    "TAIL",
    "REST",
    "MatchError",
    "match_value",
    "match_iterable",
    "match_dict"
]



import inspect
from collections.abc import Callable as ACallable
from collections.abc import Iterable, Mapping
from enum import Enum
from itertools import zip_longest
from typing import Any, Callable, Generic
from typing import Pattern as RegexPattern
from typing import Tuple, TypeVar

from pyrsistent import PVector, ny, pvector, v

from Aspidites._vendor.pampy.helpers import (BoxedArgs, HeadType, PaddedValue,
                                             TailType, UnderscoreType,
                                             get_extra,
                                             get_lambda_args_error_msg,
                                             get_real_type, is_dataclass,
                                             is_generic, is_newtype,
                                             is_typing_stuff, is_union,
                                             pairwise, peek)
from Aspidites.monads import Undefined

T = TypeVar('T')
ANY = ny
HEAD = HeadType()
REST = TAIL = TailType()


def run(action, var):
    if callable(action):
        if isinstance(var, Iterable):
            try:
                return action(*var)
            except TypeError as err:
                raise MatchError(get_lambda_args_error_msg(action, var, err))
        elif isinstance(var, BoxedArgs):
            return action(var.get())
        else:
            return action(var)
    else:
        return action


def match_value(pattern, value) -> Tuple[bool, PVector]:
    if value is PaddedValue:
        return False, v()
    elif is_typing_stuff(pattern):
        return match_typing_stuff(pattern, value)
    elif isinstance(pattern, (int, float, str, bool, Enum)):
        eq = pattern == value
        type_eq = type(pattern) == type(value)
        return eq and type_eq, v()
    elif pattern is None:
        return value is None, v()
    elif isinstance(pattern, type):
        if isinstance(value, pattern):
            return True, pvector([value])
    elif isinstance(pattern, (list, tuple)):
        return match_iterable(pattern, value)
    elif isinstance(pattern, dict):
        return match_dict(pattern, value)
    elif callable(pattern):
        return_value = pattern(value)

        if isinstance(return_value, bool):
            return return_value, pvector([value])
        elif isinstance(return_value, tuple) and len(return_value) == 2 \
                and isinstance(return_value[0], bool) and isinstance(return_value[1], list):
            return return_value
        else:
            raise MatchError("Warning! pattern function %s is not returning a boolean "
                             "nor a tuple of (boolean, list), but instead %s" %
                             (pattern, return_value))
    elif isinstance(pattern, RegexPattern):
        rematch = pattern.search(value)
        if rematch is not None:
            return True, pvector(rematch.groups())
    elif pattern is ANY:
        return True, pvector([value])
    elif pattern is HEAD or pattern is TAIL:
        raise MatchError("HEAD or TAIL should only be used inside an Iterable (list or tuple).")
    elif is_dataclass(pattern) and pattern.__class__ == value.__class__:
        return match_dict(pattern.__dict__, value.__dict__)
    return False, v()


def match_dict(pattern, value) -> Tuple[bool, PVector]:
    if not isinstance(value, dict) or not isinstance(pattern, dict):
        return False, v()

    total_extracted = []
    still_usable_value_keys = set(value.keys())
    still_usable_pattern_keys = set(pattern.keys())
    for pkey, pval in pattern.items():
        if pkey not in still_usable_pattern_keys:
            continue
        matched_left_and_right = False
        for vkey, vval in value.items():
            if vkey not in still_usable_value_keys:
                continue
            if pkey not in still_usable_pattern_keys:
                continue
            key_matched, key_extracted = match_value(pkey, vkey)
            if key_matched:
                value_matched, value_extracted = match_value(pval, vval)
                if value_matched:
                    total_extracted += key_extracted + value_extracted
                    matched_left_and_right = True
                    still_usable_pattern_keys.remove(pkey)
                    still_usable_value_keys.remove(vkey)
                    break
        if not matched_left_and_right:
            return False, v()
    return True, pvector(total_extracted)


def only_padded_values_follow(padded_pairs, i):
    i += 1
    while i < len(padded_pairs):
        pattern, value = padded_pairs[i]
        if pattern is not PaddedValue:
            return False
        i += 1
    return True


def match_iterable(patterns, values) -> Tuple[bool, PVector]:
    if not isinstance(patterns, Iterable) or not isinstance(values, Iterable):
        return False, v()

    total_extracted = []
    padded_pairs = list(zip_longest(patterns, values, fillvalue=PaddedValue))

    for i, (pattern, value) in enumerate(padded_pairs):
        if pattern is HEAD:
            if i != 0:
                raise MatchError("HEAD can only be in first position of a pattern.")
            else:
                if value is PaddedValue:
                    return False, v()
                else:
                    total_extracted += [value]
        elif pattern is TAIL:
            if not only_padded_values_follow(padded_pairs, i):
                raise MatchError("TAIL must me in last position of the pattern.")
            else:
                tail = [value for (pattern, value) in padded_pairs[i:] if value is not PaddedValue]
                total_extracted.append(tail)
                break
        else:
            matched, extracted = match_value(pattern, value)
            if not matched:
                return False, v()
            else:
                total_extracted += extracted
    return True, pvector(total_extracted)


def match_typing_stuff(pattern, value) -> Tuple[bool, PVector]:
    if pattern == Any:
        return match_value(ANY, value)
    elif is_union(pattern):
        for subpattern in pattern.__args__:
            is_matched, extracted = match_value(subpattern, value)
            if is_matched:
                return True, extracted
        else:
            return False, v()
    elif is_newtype(pattern):
        return match_value(pattern.__supertype__, value)
    elif is_generic(pattern):
        return match_generic(pattern, value)
    else:
        return False, v()


def match_generic(pattern: Generic[T], value) -> Tuple[bool, PVector]:
    if get_extra(pattern) == type:       # Type[int] for example
        real_value = None
        if is_newtype(value):
            real_value = value
            value = get_real_type(value)
        if not inspect.isclass(value):
            return False, pvector([])

        type_ = pattern.__args__[0]
        if type_ == Any:
            return True, pvector([real_value or value])
        if is_newtype(type_):   # NewType case
            type_ = get_real_type(type_)

        if issubclass(value, type_):
            return True, pvector([real_value or value])
        else:
            return False, pvector([])

    elif get_extra(pattern) == ACallable:
        if callable(value):
            spec = inspect.getfullargspec(value)
            annotations = spec.annotations
            artgtypes = [annotations.get(arg, Any) for arg in spec.args]
            ret_type = annotations.get('return', Any)
            if pattern == Callable[[*artgtypes], ret_type]:
                return True, pvector([value])
            else:
                return False, pvector([])
        else:
            return False, pvector([])

    elif get_extra(pattern) == tuple:
        return match_value(pattern.__args__, value)

    elif issubclass(get_extra(pattern), Mapping):
        type_matched, _captured = match_value(get_extra(pattern), value)
        if not type_matched:
            return False, pvector([])
        k_type, v_type = pattern.__args__

        key_example = peek(value)
        key_matched, _captured = match_value(k_type, key_example)
        if not key_matched:
            return False, pvector([])

        value_matched, _captured = match_value(v_type, value[key_example])
        if not value_matched:
            return False, pvector([])
        else:
            return True, pvector([value])

    elif issubclass(get_extra(pattern), Iterable):
        type_matched, _captured = match_value(get_extra(pattern), value)
        if not type_matched:
            return False, pvector([])
        v_type, = pattern.__args__
        v = peek(value)
        value_matched, _captured = match_value(v_type, v)
        if not value_matched:
            return False, pvector([])
        else:
            return True, pvector([value])
    else:
        return False, pvector([])


def match(var, *args, **kwargs):  # , default=NoDefault, strict=True
    """
    Match `var` against a number of potential patterns.

    Example usage:
    ```
    match(x,
        3,              "this matches the number 3",
        int,            "matches any integer",
        (str, int),     lambda a, b: "a tuple (a, b) you can use in a function",
        [1, 2, ANY],      "any list of 3 elements that begins with [1, 2]",
        {'x': ANY},       "any dict with a key 'x' and any value associated",
        ANY,              "anything else"
    )
    ```

    :param var: The variable to test patterns against.
    :param args: Alternating patterns and actions. There must be an action for every pattern specified.
                 Patterns can take many forms, see README.md for examples.
                 Actions can be either a literal value or a callable which will be called with the arguments that were
                    matched in corresponding pattern.
    :param default: If `default` is specified then it will be returned if none of the patterns match.
                    If `default` is unspecified then a `MatchError` will be thrown instead depending on the setting of strict.
    :return: The result of the action which corresponds to the first matching pattern.
    """
    if len(args) % 2 != 0:
        raise MatchError("Every guard must have an action.")

    default = False if 'default' not in kwargs.keys() else kwargs['default']
    strict = True if 'strict' not in kwargs.keys() else kwargs['strict']

    pairs = list(pairwise(args))
    patterns = [patt for (patt, action) in pairs]

    for patt, action in pairs:
        matched_as_value, args = match_value(patt, var)

        if matched_as_value:
            lambda_args = args if len(args) > 0 else BoxedArgs(var)
            return run(action, lambda_args)

    if default is False and strict:
        if ANY not in patterns:
            return Undefined()
    else:
        return default


class MatchError(Exception):
    def __init__(self, msg):
        super().__init__(msg)



import ast
from .._compat import IS_PY2 as IS_PY2, IS_PY3 as IS_PY3, IS_PY34_OR_GREATER as IS_PY34_OR_GREATER, IS_PY35_OR_GREATER as IS_PY35_OR_GREATER
from typing import Any

IOPERATOR_TO_STR: Any
ALLOWED_FUNC_NAMES: Any
FORBIDDEN_FUNC_NAMES: Any

def copy_locations(new_node, old_node) -> None: ...

class PrintInfo:
    print_used: bool
    printed_used: bool
    def __init__(self) -> None: ...
    def new_print_scope(self) -> None: ...

class RestrictingNodeTransformer(ast.NodeTransformer):
    errors: Any
    warnings: Any
    used_names: Any
    print_info: Any
    def __init__(self, errors: Any | None = ..., warnings: Any | None = ..., used_names: Any | None = ...) -> None: ...
    def gen_tmp_name(self): ...
    def error(self, node, info) -> None: ...
    def warn(self, node, info) -> None: ...
    def guard_iter(self, node): ...
    def is_starred(self, ob): ...
    def gen_unpack_spec(self, tpl): ...
    def protect_unpack_sequence(self, target, value): ...
    def gen_unpack_wrapper(self, node, target, ctx: str = ...): ...
    def gen_none_node(self): ...
    def gen_lambda(self, args, body): ...
    def gen_del_stmt(self, name_to_del): ...
    def transform_slice(self, slice_): ...
    def check_name(self, node, name, allow_magic_methods: bool = ...) -> None: ...
    def check_function_argument_names(self, node) -> None: ...
    def check_import_names(self, node): ...
    def inject_print_collector(self, node, position: int = ...) -> None: ...
    def gen_attr_check(self, node, attr_name): ...
    def generic_visit(self, node): ...
    def not_allowed(self, node) -> None: ...
    def node_contents_visit(self, node): ...
    def visit_Num(self, node): ...
    def visit_Str(self, node): ...
    def visit_Bytes(self, node): ...
    def visit_List(self, node): ...
    def visit_Tuple(self, node): ...
    def visit_Set(self, node): ...
    def visit_Dict(self, node): ...
    def visit_FormattedValue(self, node): ...
    def visit_JoinedStr(self, node): ...
    def visit_Constant(self, node): ...
    def visit_Ellipsis(self, node): ...
    def visit_NameConstant(self, node): ...
    def visit_Name(self, node): ...
    def visit_Load(self, node): ...
    def visit_Store(self, node): ...
    def visit_Del(self, node): ...
    def visit_Starred(self, node): ...
    def visit_Expression(self, node): ...
    def visit_Expr(self, node): ...
    def visit_UnaryOp(self, node): ...
    def visit_UAdd(self, node): ...
    def visit_USub(self, node): ...
    def visit_Not(self, node): ...
    def visit_Invert(self, node): ...
    def visit_BinOp(self, node): ...
    def visit_Add(self, node): ...
    def visit_Sub(self, node): ...
    def visit_Mult(self, node): ...
    def visit_Div(self, node): ...
    def visit_FloorDiv(self, node): ...
    def visit_Mod(self, node): ...
    def visit_Pow(self, node): ...
    def visit_LShift(self, node): ...
    def visit_RShift(self, node): ...
    def visit_BitOr(self, node): ...
    def visit_BitXor(self, node): ...
    def visit_BitAnd(self, node): ...
    def visit_MatMult(self, node) -> None: ...
    def visit_BoolOp(self, node): ...
    def visit_And(self, node): ...
    def visit_Or(self, node): ...
    def visit_Compare(self, node): ...
    def visit_Eq(self, node): ...
    def visit_NotEq(self, node): ...
    def visit_Lt(self, node): ...
    def visit_LtE(self, node): ...
    def visit_Gt(self, node): ...
    def visit_GtE(self, node): ...
    def visit_Is(self, node): ...
    def visit_IsNot(self, node): ...
    def visit_In(self, node): ...
    def visit_NotIn(self, node): ...
    def visit_Call(self, node): ...
    def visit_keyword(self, node): ...
    def visit_IfExp(self, node): ...
    def visit_Attribute(self, node): ...
    def visit_Subscript(self, node): ...
    def visit_Index(self, node): ...
    def visit_Slice(self, node): ...
    def visit_ExtSlice(self, node): ...
    def visit_ListComp(self, node): ...
    def visit_SetComp(self, node): ...
    def visit_GeneratorExp(self, node): ...
    def visit_DictComp(self, node): ...
    def visit_comprehension(self, node): ...
    def visit_Assign(self, node): ...
    def visit_AugAssign(self, node): ...
    def visit_Print(self, node): ...
    def visit_Raise(self, node): ...
    def visit_Assert(self, node): ...
    def visit_Delete(self, node): ...
    def visit_Pass(self, node): ...
    def visit_Import(self, node): ...
    def visit_ImportFrom(self, node): ...
    def visit_alias(self, node): ...
    def visit_Exec(self, node) -> None: ...
    def visit_If(self, node): ...
    def visit_For(self, node): ...
    def visit_While(self, node): ...
    def visit_Break(self, node): ...
    def visit_Continue(self, node): ...
    def visit_Try(self, node): ...
    def visit_TryFinally(self, node): ...
    def visit_TryExcept(self, node): ...
    def visit_ExceptHandler(self, node): ...
    def visit_With(self, node): ...
    def visit_withitem(self, node): ...
    def visit_FunctionDef(self, node): ...
    def visit_Lambda(self, node): ...
    def visit_arguments(self, node): ...
    def visit_arg(self, node): ...
    def visit_Return(self, node): ...
    def visit_Yield(self, node): ...
    def visit_YieldFrom(self, node): ...
    def visit_Global(self, node): ...
    def visit_Nonlocal(self, node) -> None: ...
    def visit_ClassDef(self, node): ...
    def visit_Module(self, node): ...
    def visit_Param(self, node): ...
    def visit_AsyncFunctionDef(self, node) -> None: ...
    def visit_Await(self, node) -> None: ...
    def visit_AsyncFor(self, node) -> None: ...
    def visit_AsyncWith(self, node) -> None: ...
    def visit_NamedExpr(self, node): ...




# Aspidites
# Copyright (C) 2021 Ross J. Duff

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import inspect
from math import inf, isinf
from contextlib import suppress
from _warnings import warn

from pyrsistent import v

from ._vendor.contracts import ContractNotRespected, contract
from ._vendor.fn import apply
from ._vendor.fn.underscore import ArityError
from .api import ContractBreachWarning, create_warning


# noinspection PyPep8Naming
def SafeDiv(a, b):
    """IEEE 754-1985 evaluates an expression and replaces indeterminate forms with Undefined instances"""
    if b == 0 or (isinf(a) and isinf(b)):
        stack = inspect.stack()
        w = create_warning(
            stack[0][3],
            [a, b],
            {},
            stack,
            ZeroDivisionError("Division by zero is Undefined; this behavior diverges from IEEE 754-1985.")
        )
        warn(w, category=RuntimeWarning)
        return Undefined(SafeDiv, a=a, b=b)
    return a / b


# noinspection PyPep8Naming
def SafeMod(a, b):
    """IEEE 754-1985 evaluates an expression and replaces indeterminate forms with Undefined instances"""
    if isinf(a) or b == 0:
        stack = inspect.stack()
        w = create_warning(
            stack[0][3],
            [a, b],
            {},
            stack,
            ZeroDivisionError("Modulus by zero is Undefined; this behavior diverges from IEEE 754-1985.")
        )
        warn(w, category=RuntimeWarning)
        return Undefined(SafeMod, a=a, b=b)
    return a % b


# noinspection PyPep8Naming
def SafeExp(a, b):
    if ((a == 0 and b == 0) or (isinf(a) and b == 0) or (isinf(b) and a == 0)):  # 0**0, inf**0, 0**inf
        stack = inspect.stack()
        w = create_warning(
            stack[0][3], [a, b], {}, stack,
            ArithmeticError("%s**%s" % (str(a), str(b),) + " == Undefined; this behavior diverges from IEEE 754-1985.")
        )
        warn(w, category=RuntimeWarning)
        return Undefined(SafeExp, a=a, b=b)
    try:
        return a**b
    except OverflowError:
        return inf  # just a really big number on most systems


class Maybe:  # this should allow us to avoid passing around stacks and frames
    """Sandboxes a Surely call and handles ContractNotRespected by returning Undefined"""

    __slots__ = v("_func", "_args", "_kwargs", "__instance__")

    def __init__(self, func, *args, **kwargs):
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self.__instance__ = Undefined()

    def __repr__(self):
        maybe = self.__class__.__name__
        inst = self.__instance__
        inst_undef = inst == Undefined()
        debug = (" -> %s" % Undefined() if inst_undef else " -> %s" % str(inst))
        fname = str(self._func.__name__)
        args = str(self._args).strip("()")
        kwargs = [str(k) + " = " + str(v) for k, v in self._kwargs.items()]
        return maybe + "(" + ", ".join([fname, args, *kwargs]) + ")" + debug

    def __invert__(self):
        return ~self.__instance__

    def __neg__(self):
        return -self.__instance__

    @property
    def func(self):
        return self._func

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs

    def __call__(self, warn_undefined=True):
        try:
            with suppress(ValueError):
                val = apply(self.func, self.args, self.kwargs)
            with suppress(UnboundLocalError):
                self.__instance__ = Surely(val)
                # SURELY #
                return self.__instance__
            self.__instance__ = Undefined(self.func, self.args, self.kwargs)
        except (ContractNotRespected, ArityError, ZeroDivisionError, Exception) as e:
            if warn_undefined:
                stack = inspect.stack(1)
                w = create_warning(self.func, self.args, self.kwargs, stack, e)
                warn(w, category=ContractBreachWarning if isinstance(e, ContractNotRespected) else RuntimeWarning)
            # UNDEFINED #
            self.__instance__ = Undefined(self.func, self.args, self.kwargs)
            return self.__instance__


class Undefined:
    """A monad for a failed programmatic unit; like NoneType but hashable.
    Falsy singleton"""

    __slots__ = v("__weakref__", "__instance__")
    __instance = None

    def __hash__(self):
        return self

    def __eq__(self, other):
        return self.__hash__ == other.__hash__

    def __add__(self, other):
        return other

    def __sub__(self, other):
        return -other

    def __neg__(self):
        return self

    def __invert__(self):
        return self

    def __mul__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __str__(self):
        return str(self.__repr__())

    def __float__(self):
        return float()

    def __complex__(self):
        return complex()

    def __oct__(self):
        return oct(0)

    def __index__(self):
        return 0

    def __len__(self):
        # Undefined has 0 elements
        return 0

    def __repr__(self):
        return self.__class__.__name__

    def __nonzero__(self):
        return False

    def __call__(self, *args, **kwargs):
        return self.__new__(self.__class__, *args, **kwargs)

    def __new__(mcs, *args, **kwargs):
        if mcs.__instance is None:
            mcs.__instance = super(Undefined, mcs).__new__(mcs, *args, **kwargs)
            mcs.__instance__ = mcs.__instance
        return mcs.__instance__  # instance descriptor from __slots__ -> actual instance


class Surely:
    """A monad for a successful programmatic unit
    Truthy, defers to an instance of a successful computation"""

    __slots__ = v(
        "__weakref__", "__instance__" "__str__", "__int__", "__float__", "__complex__"
    )

    # def __try_except_undefined__(self, other=None, call=None):
    #     if call:
    #         # noinspection PyComparisonWithNone
    #         return (
    #             call(self.__instance__)
    #             if other == None
    #             else call(self.__instance__, other)
    #         )
    #     else:
    #         return self.__instance__
    #
    # def __hash__(self):
    #     return self.__try_except_undefined__(call=hash)
    #
    # def __eq__(self, other):
    #     return self.__try_except_undefined__(other, op.eq)
    #
    # def __add__(self, other):
    #     return self.__try_except_undefined__(other, op.add)
    #
    # def __sub__(self, other):
    #     return self.__try_except_undefined__(other, op.sub)
    #
    # def __neg__(self):
    #     return self.__try_except_undefined__(call=op.neg)
    #
    # def __invert__(self):
    #     return self.__try_except_undefined__(call=op.invert)
    #
    # def __mul__(self, other):
    #     return self.__try_except_undefined__(other, op.mul)
    #
    # def __truediv__(self, other):
    #     return self.__try_except_undefined__(other, op.truediv)
    #
    # def __floordiv__(self, other):
    #     return self.__try_except_undefined__(other, op.floordiv)
    #
    # def __oct__(self):
    #     return self.__try_except_undefined__(call=oct)
    #
    # def __nonzero__(self):
    #     return self.__try_except_undefined__(call=bool)
    #
    # @classmethod
    # def __call__(cls, *args, **kwargs):
    #     return cls.__instance__

    def __new__(cls, instance__=Undefined(), *args, **kwargs):
        cls.__instance__ = instance__
        return cls.__instance__



__version__ = '0.12.0'
__title__ = 'Aspidites'
__author__ = 'Ross J. Duff'
__license__ = "GPL"
__description__ = (
    f"""
    {__title__} v{__version__} is the reference Woma programming language compiler.
    Copyright (C) 2021  {__author__}
    This program comes with ABSOLUTELY NO WARRANTY; 
    This is free software, and you are welcome to redistribute it
    under the conditions of the {__license__}v3 found in the LICENSE file.
    """
)
__mimetype__ = 'text/woma'



from ..interface import Contract, ContractNotRespected
from ..syntax import add_contract, W, Literal


class Any(Contract):
    """ Always true. """

    def __init__(self, where=None):
        Contract.__init__(self, where)

    def __repr__(self):
        return 'Any()'

    def __str__(self):
        return '*'

    def check_contract(self, context, value, silent):
        pass

    @staticmethod
    def parse_action(s, loc, tokens):  # @UnusedVariable
        return Any(W(s, loc))


class Never(Contract):
    """ A contract that does not match anything. Useful for debugging. """

    def __init__(self, where=None):
        Contract.__init__(self, where)

    def __repr__(self):
        return 'Never()'

    def __str__(self):
        return '#'

    def check_contract(self, context, value, silent):
        raise ContractNotRespected(self, 'No value can match this',
                                   value, context)

    @staticmethod
    def parse_action(s, loc, tokens):  # @UnusedVariable
        return Never(W(s, loc))


add_contract(Literal('*').setParseAction(Any.parse_action))
add_contract(Literal('#').setParseAction(Never.parse_action))



from Aspidites._vendor.RestrictedPython import compile_restricted_function
from Aspidites._vendor.RestrictedPython import PrintCollector
from Aspidites._vendor.RestrictedPython import safe_builtins
from Aspidites._vendor._compat import IS_PY38_OR_GREATER
from types import FunctionType


def test_compile_restricted_function():
    p = ''
    body = """
print("Hello World!")
return printed
"""
    name = "hello_world"
    global_symbols = []

    result = compile_restricted_function(
        p,  # parameters
        body,
        name,
        filename='<string>',
        globalize=global_symbols
    )

    assert result.code is not None
    assert result.errors == ()

    safe_globals = {
        '__name__': 'script',
        '_getattr_': getattr,
        '_print_': PrintCollector,
        '__builtins__': safe_builtins,
    }
    safe_locals = {}
    exec(result.code, safe_globals, safe_locals)
    hello_world = safe_locals['hello_world']
    assert type(hello_world) == FunctionType
    assert hello_world() == 'Hello World!\n'


def test_compile_restricted_function_func_wrapped():
    p = ''
    body = """
print("Hello World!")
return printed
"""
    name = "hello_world"
    global_symbols = []

    result = compile_restricted_function(
        p,  # parameters
        body,
        name,
        filename='<string>',
        globalize=global_symbols
    )

    assert result.code is not None
    assert result.errors == ()
    safe_globals = {
        '__name__': 'script',
        '_getattr_': getattr,
        '_print_': PrintCollector,
        '__builtins__': safe_builtins,
    }

    func = FunctionType(result.code, safe_globals)
    func()
    assert 'hello_world' in safe_globals
    hello_world = safe_globals['hello_world']
    assert hello_world() == 'Hello World!\n'


def test_compile_restricted_function_with_arguments():
    p = 'input1, input2'
    body = """
print(input1 + input2)
return printed
"""
    name = "hello_world"
    global_symbols = []

    result = compile_restricted_function(
        p,  # parameters
        body,
        name,
        filename='<string>',
        globalize=global_symbols
    )

    assert result.code is not None
    assert result.errors == ()

    safe_globals = {
        '__name__': 'script',
        '_getattr_': getattr,
        '_print_': PrintCollector,
        '__builtins__': safe_builtins,
    }
    safe_locals = {}
    exec(result.code, safe_globals, safe_locals)
    hello_world = safe_locals['hello_world']
    assert type(hello_world) == FunctionType
    assert hello_world('Hello ', 'World!') == 'Hello World!\n'


def test_compile_restricted_function_can_access_global_variables():
    p = ''
    body = """
print(input)
return printed
"""
    name = "hello_world"
    global_symbols = ['input']

    result = compile_restricted_function(
        p,  # parameters
        body,
        name,
        filename='<string>',
        globalize=global_symbols
    )

    assert result.code is not None
    assert result.errors == ()

    safe_globals = {
        '__name__': 'script',
        '_getattr_': getattr,
        'input': 'Hello World!',
        '_print_': PrintCollector,
        '__builtins__': safe_builtins,
    }
    safe_locals = {}
    exec(result.code, safe_globals, safe_locals)
    hello_world = safe_locals['hello_world']
    assert type(hello_world) == FunctionType
    assert hello_world() == 'Hello World!\n'


def test_compile_restricted_function_pretends_the_code_is_executed_in_a_global_scope():  # NOQA: E501
    p = ''
    body = """output = output + 'bar'"""
    name = "hello_world"
    global_symbols = ['output']

    result = compile_restricted_function(
        p,  # parameters
        body,
        name,
        filename='<string>',
        globalize=global_symbols
    )

    assert result.code is not None
    assert result.errors == ()

    safe_globals = {
        '__name__': 'script',
        'output': 'foo',
        '__builtins__': {},
    }
    safe_locals = {}
    exec(result.code, safe_globals, safe_locals)
    hello_world = safe_locals['hello_world']
    assert type(hello_world) == FunctionType
    hello_world()
    assert safe_globals['output'] == 'foobar'


def test_compile_restricted_function_allows_invalid_python_identifiers_as_function_name():  # NOQA: E501
    p = ''
    body = """output = output + 'bar'"""
    name = "<foo>.bar.__baz__"
    global_symbols = ['output']

    result = compile_restricted_function(
        p,  # parameters
        body,
        name,
        filename='<string>',
        globalize=global_symbols
    )

    assert result.code is not None
    assert result.errors == ()

    safe_globals = {
        '__name__': 'script',
        'output': 'foo',
        '__builtins__': {},
    }
    safe_locals = {}
    exec(result.code, safe_globals, safe_locals)
    generated_function = tuple(safe_locals.values())[0]
    assert type(generated_function) == FunctionType
    generated_function()
    assert safe_globals['output'] == 'foobar'


def test_compile_restricted_function_handle_SyntaxError():
    p = ''
    body = """a("""
    name = "broken"

    result = compile_restricted_function(
        p,  # parameters
        body,
        name,
    )

    assert result.code is None
    assert result.errors == (
        "Line 1: SyntaxError: unexpected EOF while parsing at statement: 'a('",
    )


def test_compile_restricted_function_invalid_syntax():
    p = ''
    body = '1=1'
    name = 'broken'

    result = compile_restricted_function(
        p,  # parameters
        body,
        name,
    )

    assert result.code is None
    assert len(result.errors) == 1
    error_msg = result.errors[0]

    if IS_PY38_OR_GREATER:
        assert error_msg.startswith(
            "Line 1: SyntaxError: cannot assign to literal at statement:"
        )
    else:
        assert error_msg.startswith(
            "Line 1: SyntaxError: can't assign to literal at statement:"
        )



import operator
import itertools
from Aspidites._vendor.fn import _, F
from Aspidites._vendor.fn import op


def test_unfold():
    doubler = op.unfold(lambda x: (x * 2, x * 2))
    assert list(itertools.islice(doubler(10), 0, 10)) == [20, 40, 80, 160, 320, 640,
                                                          1280, 2560, 5120, 10240]


def test_currying():
    def add(first):
        def add(second):
            return first + second

        return add

    assert 1 == op.curry(add, 0, 1)


def test_apply():
     assert 10, op.apply(operator.add, [2, 8])


def test_flip():
     assert 10 == op.flip(operator.sub)(2, 12)
     assert -10 == op.flip(op.flip(operator.sub))(2, 12)
    # flipping of flipped function should use optimization
     assert bool(operator.sub is op.flip(op.flip(operator.sub))) is True


def test_flip_with_shortcut():
     assert 10 == op.flip(_ - _)(2, 12)


def test_zipwith():
    zipper = op.zipwith(operator.add)
    assert [10, 11, 12] == list(zipper([0, 1, 2], itertools.repeat(10)))

    zipper = op.zipwith(_ + _)
    assert [10, 11, 12] == list(zipper([0, 1, 2], itertools.repeat(10)))

    zipper = F() << list << op.zipwith(_ + _)
    assert [10, 11, 12] == zipper([0, 1, 2], itertools.repeat(10))


def test_foldl():
    assert 10 == op.foldl(operator.add)([0, 1, 2, 3, 4])
    assert 20 == op.foldl(operator.add, 10)([0, 1, 2, 3, 4])
    assert 20 == op.foldl(operator.add, 10)(range(5))
    assert 10 == op.foldl(_ + _)(range(5))


def test_foldr():
    summer = op.foldr(operator.add)
    assert 10 == op.foldr(operator.add)([0, 1, 2, 3, 4])
    assert 20 == op.foldr(operator.add, 10)([0, 1, 2, 3, 4])
    assert 20 == op.foldr(operator.add, 10)(range(5))
    # specific case for right-side folding
    assert 100 == op.foldr(op.call, 0)([lambda s: s ** 2, lambda k: k + 10])



from .reserved import *


def cvt_arith_expr(tks):
    expr = "".join((str(t) for t in tks))
    if "/" in expr:
        return "Maybe(SafeDiv, " + expr.replace("/", ", ") + ")"
    elif "%" in expr:
        return "Maybe(SafeMod, " + expr.replace("%", ", ") + ")"
    elif "**" in expr:
        return "Maybe(SafeExp, " + expr.replace("**", ", ") + ")"


def cvt_pragma(tks):
    t: list = tks.asList()
    return '\n'.join(t) + '\n'


def cvt_int(t):
    return int(t[0])


def cvt_real(t):
    return float(t[0])


def cvt_tuple(t):
    return "(" + ", ".join(t.asList()) + ")"


def cvt_comment_line(s, loc, t):
    return "# comment_line %s:" % (len([c for c in s[:loc] if c == "\n"]) + 1) + t[0]


def cvt_dict(t):
    t = t.asList()
    for i, v in enumerate(t):
        key, val = v
        if isinstance(key, str):  # string keys only
            t[i] = f"{key}: {val}"
        else:  # integer key
            t[i] = f"{key}: {val}"
    t = f'{", ".join(t)}'
    return f"pmap({{{t}}})"


def cvt_list(t):
    t = t.asList()
    for i, v in enumerate(t):
        if isinstance(v, str):  # string keys only
            t[i] = v
        else:  # integer key
            t[i] = int(v)
    t = f'{", ".join(t)}'
    return f"pvector([{t}])"


def cvt_set(t):
    t = t.asList()
    for i, v in enumerate(t):
        if isinstance(v, str):  # string keys only
            t[i] = v
        else:  # integer key
            t[i] = int(v)
    t = f'{", ".join(t)}'
    return f"pset({{{t}}})"


def cvt_contract_assign(t):
    t = swap_val_to_idx(list(t), ":", 1)
    t[2], t[4] = t[4], t[2]
    return " ".join((str(t) for t in t))


def cvt_contract_define(t):
    t[0], t[1] = t[1], t[0]
    t[1] = "'" + t[1] + "'"
    args = f"({', '.join(t[1:])})"
    t = t[0] + args
    return "".join(t)


def swap_val_to_idx(lst: list, val, idx: int) -> list:
    val_idx = lst.index(val)
    if val_idx == idx:
        pass  # maybe error?
    lst[val_idx], lst[idx] = lst[idx], lst[val_idx]
    return lst


def cvt_clos_call(t):
    return "Maybe" + t[0][1] + t[0][0] + sep + sep.join(t[0][2:-1]) + t[0][-1]


def cvt_func_call(t):
    return "Maybe" + t[0][1] + t[0][0] + sep + sep.join(t[0][2:-1]) + t[0][-1] + lit_lparen + lit_rparen


from ..interface import Contract, ContractNotRespected, describe_value
from ..syntax import (Combine, Word, W, alphas, alphanums, oneOf,
                      ParseException, ZeroOrMore, S, rvalue,
                      delimitedList, Optional)
from pyparsing import ParseFatalException


class Extension(Contract):
    registrar = {}

    def __init__(self, identifier, where=None, args=tuple(), kwargs=None):
        assert identifier in Extension.registrar
        self.contract = Extension.registrar[identifier]
        self.identifier = identifier
        self.args = args
        self.kwargs = kwargs or {}
        Contract.__init__(self, where)

    def __str__(self):
        inside = []
        if self.args:
            inside.extend(map(str, self.args))
            
        if self.kwargs:
            ks = sorted(self.kwargs)
            inside.extend(["%s=%s" % (k, self.kwargs[k]) for k in ks])

        s = self.identifier
    
        if inside:
            return self.identifier + "(" + ",".join(inside) + ")"
        else:
            return s

    def __repr__(self):
        if self.args or self.kwargs:
            return ("Extension(%r, args=%r, kwargs=%r)" %
                    (self.identifier, self.args, self.kwargs))

        return "Extension(%r)" % self.identifier

    def check_contract(self, context, value, silent):
        context['args'] = tuple(a.eval(context) for a in self.args)
        context['kwargs'] = dict((k, v.eval(context)) for
                                 k, v in self.kwargs.items())

        # noinspection PyProtectedMember
        self.contract._check_contract(context, value, silent)

    @staticmethod
    def parse_action(s, loc, tokens):
        identifier = tokens[0]
        args = tuple()
        kwargs = {}

        if len(tokens) == 2:
            args, kwargs = tokens[1]
            args = tuple(args)

        if not identifier in Extension.registrar:
            raise ParseException('Unknown extension contract %r' % identifier)
        
        # from contracts.library.separate_context import SeparateContext
        
        contract_ext = Extension.registrar[identifier]
        
        if isinstance(contract_ext, CheckCallable):
            callable_thing = contract_ext.callable 
         
            test_args = ('value',) + args
            from Aspidites._vendor.contracts.inspection import check_callable_accepts_these_arguments, InvalidArgs
         
            try:
                check_callable_accepts_these_arguments(callable_thing, test_args, kwargs)
             
            except InvalidArgs as e:
                msg = 'The callable %s cannot accept these arguments ' % callable_thing
                msg += 'args = %s, kwargs = %s ' % (test_args, kwargs)
                msg += '%s' % e
                raise ParseFatalException(msg)

        where = W(s, loc)
        return Extension(identifier, where, args, kwargs)

    # We want to be pickable so we do not save self.contract
    # which might point to a lambda
    def __getstate__(self):
        return {'identifier': self.identifier,
                'args': self.args,
                'kwargs': self.kwargs}

    def __setstate__(self, d):
        self.identifier = d['identifier']
        self.contract = Extension.registrar[self.identifier]
        self.args = d['args']
        self.kwargs = d['kwargs']


class CheckCallable(Contract):

    def __init__(self, callable):
        self.callable = callable
        Contract.__init__(self, where=None)

    def check_contract(self, context, value, silent):
        allowed = (ValueError, AssertionError)
        args = context.get('args', tuple())
        kwargs = context.get('kwargs', {})
        try:
            result = self.callable(value, *args, **kwargs)
        except allowed as e:  # failed
            raise ContractNotRespected(self, str(e), value, context)

        if result in [None, True]:
            # passed
            pass
        elif result == False:
            msg = ('Value does not pass criteria of %s() (module: %s).' %
                   (get_callable_name(self.callable), 
                    get_callable_module(self.callable)))
            raise ContractNotRespected(self, msg, value, context)
        else:
            msg = ('I expect that %r returns either True, False, None; or '
                   'raises a ValueError exception. Instead, I got %s.' %
                   (self.callable, describe_value(value)))
            raise ValueError(msg)

    def __repr__(self):
        """ Note: this contract is not representable, but anyway it is
            only used by Extension, which serializes using the identifier. """
        return 'CheckCallable(%r)' % self.callable

    def __str__(self):
        """ Note: this contract is not representable, but anyway it is only
            used by Extension, which serializes using the identifier. """
        return get_callable_name(callable)


def get_callable_name(c):
    """ Get a displayable name for the callable even if __name__
        is not available. """
    try:
        return c.__name__ + '()'
    except AttributeError:
        return str(c)


def get_callable_module(c):
    try:
        return c.__module__
    except AttributeError:
        return '(No __module__ attr)'


def describe_callable(c):
    return get_callable_name(c) + ' module: %s' % get_callable_module(c)


class CheckCallableWithSelf(Contract):

    def __init__(self, callable):  # @ReservedAssignment
        self.callable = callable
        Contract.__init__(self, where=None)

    def check_contract(self, context, value, silent):
        args = context.get('args', tuple())
        kwargs = context.get('kwargs', {})

        if not 'self' in context:
            msg = ('You can only call this contract in the context of '
                   ' a function call to a regular method.')
            raise ContractNotRespected(self, msg, value, context)

        args = (context['self'], value) + args
        allowed = (ValueError, AssertionError)
        try:
            result = self.callable(*args, **kwargs)
        except allowed as e:  # failed
            raise ContractNotRespected(self, str(e), value, context)

        if result in [None, True]:
            # passed
            pass
        elif result == False:
            msg = ('Value does not pass criteria of %s.' %
                   describe_callable(self.callable))
            raise ContractNotRespected(self, msg, value, context)
        else:
            msg = ('I expect that %r returns either True, False, None; or '
                   'raises a ValueError exception. Instead, I got %s.' %
                   (self.callable, describe_value(value)))
            raise ValueError(msg)

    def __repr__(self):
        """ Note: this contract is not representable, but anyway it is only
            used by Extension, which serializes using the identifier. """
        return 'CheckCallableWithSelf(%r)' % self.callable

    def __str__(self):
        """ Note: this contract is not representable, but anyway it is only
            used by Extension, which serializes using the identifier. """
        return 'function %s()' % get_callable_name(self.callable)


w = Word('_' + alphanums)
arg = rvalue.copy()

kwarg = w + ZeroOrMore(' ') + S('=') + ZeroOrMore(' ') + rvalue
kwarg.setParseAction(lambda s, loc, tokens: {tokens[0]: tokens[1]})


def build_args_kwargs(s, loc, tokens):
    return (tuple(t for t in tokens if not isinstance(t, dict)),
            dict((k, v) for t in tokens if isinstance(t, dict)
                 for k, v in t.items()))


arglist = delimitedList(kwarg | arg)
arglist.setParseAction(build_args_kwargs)

identifier_expression = (Combine(oneOf(list(alphas)) + Word('_' + alphanums)) +
                         Optional(S('(') + arglist + S(')')))


identifier_contract = identifier_expression.copy().setParseAction(
    Extension.parse_action)



from . import syntax_fail, good, fail


### Strings
good('str', 'ciao')
good('string', 'ciao', exact=False)
fail('str', list('ciao'))  # sequences of chars are not str
# Can specify the length
good('str[N]', '')
good('str[1]', 'a')
good('str[2]', 'ab')
good('str[>0]', 'ab')
fail('str[>0]', '')
good('str[N],N>3', 'ciao')
fail('str[N],N>3', 'cia')
# Cannot specify anything on the content
syntax_fail('str(*)')



from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites._vendor.RestrictedPython import safe_globals
from Aspidites._vendor.RestrictedPython.Eval import default_guarded_getiter
from Aspidites._vendor.RestrictedPython.Guards import guarded_iter_unpack_sequence

import pytest


ITERATE_OVER_DICT_ITEMS = """
d = {'a': 'b'}
for k, v in d.items():
    pass
"""


def test_iterate_over_dict_items_plain():
    glb = {}
    result = compile_restricted_exec(ITERATE_OVER_DICT_ITEMS)
    assert result.code is not None
    assert result.errors == ()
    with pytest.raises(NameError) as excinfo:
        exec(result.code, glb, None)
    assert "name '_iter_unpack_sequence_' is not defined" in str(excinfo.value)


def test_iterate_over_dict_items_safe():
    glb = safe_globals.copy()
    glb['_getiter_'] = default_guarded_getiter
    glb['_iter_unpack_sequence_'] = guarded_iter_unpack_sequence
    result = compile_restricted_exec(ITERATE_OVER_DICT_ITEMS)
    assert result.code is not None
    assert result.errors == ()
    exec(result.code, glb, None)
